from abc import abstractmethod


class DataLoggerMap(dict):

    @property
    @abstractmethod
    def key_type(self):
        raise NotImplementedError()

    def __init__(self, kwargs=None):
        if kwargs:
            super(DataLoggerMap, self).__init__(kwargs)
        else:
            super(DataLoggerMap, self).__init__()

    @classmethod
    def build(cls, *items):
        assert len(items) % 2 == 0, "The number of items must be even in order to make pairs."
        dlm = cls()

        paired_items = list()
        assert len(items) % 2 == 0, "The number of items to build pairs is not even; In order to build pairs from a list of items, math mandates an even number of items."
        for i, item in enumerate(items):
            # if i == i, then i - 1 == 0;
            # e.g. items[i - 1] is the name of the field/tag and items[i] is the value.
            if i % 2 != 0:
                name = items[i - 1]
                value = items[i]
                assert isinstance(name, dlm.key_type) and not isinstance(value, dlm.key_type), f"The pairs must be on the format: ({dlm.key_type}, (int, float, str, etc, but NOT {dlm.key_type}))"
                paired_items.append((name, value))

        for pair in paired_items:
            assert isinstance(pair, tuple)
            dlm.update({pair[0]: pair[1]})

        return dlm



import datetime
import json
import logging
import sys
import os
import time
import socket

DIR = os.path.dirname(os.path.realpath(__file__))
path = f"{os.path.dirname(DIR)}"
temp_path = [path]
temp_path.extend(sys.path)
sys.path = temp_path


import influxdb_datalogger

logger = logging.getLogger("pytest")

DATA_FILE = f"{DIR}/data-file.json"
EVENT_FILE = f"{DIR}/event-file.json"


class DbWriter(influxdb_datalogger.DatabaseWriter):
    def write_data(self, datalogger: influxdb_datalogger.DataLogger):
        dataset = datalogger.dataset
        if os.path.exists(DATA_FILE):
            # If the file exists we are probably writing more than once in a test.
            # Since DbWriter just writes to a JSON file we need to essentially append to the existing data, which is what this does.

            loaded_dataset = json.load(open(DATA_FILE))
            dataset = [*loaded_dataset, *dataset]
        json.dump(dataset, open(DATA_FILE, "w"), indent=2)


class GrafanaEventWriter(influxdb_datalogger.DatabaseWriter):
    def write_data(self, datalogger: influxdb_datalogger.DataLogger):
        events = datalogger.events
        events_to_write = list()
        for event in events:
            event: influxdb_datalogger.datalogger.Event
            marker_data = dict(time=int(event.start * 1000), timeEnd=int(event.stop * 1000), tags=event.tags_map, text=event.measurement)
            events_to_write.append(marker_data)

        if os.path.exists(EVENT_FILE):
            # If the file exists we are probably writing more than once in a test.
            # Since DbWriter just writes to a JSON file we need to essentially append to the existing data, which is what this does.

            loaded_events = json.load(open(EVENT_FILE))
            events_to_write = [*loaded_events, *events_to_write]
        json.dump(events_to_write, open(EVENT_FILE, "w"), indent=2)


class TestDatalogger:

    ############################################################################################################
    # Setup/teardown
    ############################################################################################################

    def setup_method(self):
        self.remove_data_file()
        self.remove_events_file()

    def teardown_method(self):
        self.remove_data_file()
        self.remove_events_file()

    ############################################################################################################
    # Functions
    ############################################################################################################

    def remove_data_file(self):
        try:
            os.remove(DATA_FILE)
        except:
            # Probably doesn't exist; this is fine
            pass

    def remove_events_file(self):
        try:
            os.remove(EVENT_FILE)
        except:
            # Probably doesn't exist; this is fine
            pass

    def assert_data_file_doesnt_exist(self):
        assert not os.path.exists(DATA_FILE), f"File {DATA_FILE} exists when it shouldn't at this point"

    def assert_event_file_doesnt_exist(self):
        assert not os.path.exists(EVENT_FILE), f"File {EVENT_FILE} exists when it shouldn't at this point"

    def load_dataset_from_data_file(self):
        return json.load(open(DATA_FILE))

    def load_events_from_event_file(self):
        return json.load(open(EVENT_FILE))

    ############################################################################################################
    # Tests
    ############################################################################################################

    def test_field_map(self):
        f = influxdb_datalogger.Field("duration")
        fm = influxdb_datalogger.FieldMap.build(f, 2)

        assert f in fm
        assert fm[f] == 2

    def test_field_map_fail(self):
        try:
            fm = influxdb_datalogger.FieldMap.build(influxdb_datalogger.Tag("duration"), 2)
            raise Exception("Managed to create a FieldMap that should not be possible to create")
        except:
            logger.info("Correctly failed to create a map that should not be possible to make")

    def test_tags_map(self):
        t = influxdb_datalogger.Tag("name")
        tm = influxdb_datalogger.TagMap.build(t, "testing")

        assert t in tm
        assert tm[t] == "testing"

    def test_tags_map_fail(self):
        try:
            tm = influxdb_datalogger.TagMap.build(influxdb_datalogger.Field("duration"), 2)
            raise Exception("Managed to create a TagMap that should not be possible to create")
        except:
            logger.info("Correctly failed to create a map that should not be possible to make")

    def test_block_measure(self):
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        m = influxdb_datalogger.Measurement("time-taken", f, t)
        tm = influxdb_datalogger.TagMap.build(t, "test")
        datalogger = influxdb_datalogger.DataLogger(DbWriter())
        with datalogger.measure_block(m, f, tm):
            logger.info(f"Taking measurement {m}")

        assert m in datalogger._dataset, f"Measurement {m} not in the dataset"

    def test_write_data(self):
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        m = influxdb_datalogger.Measurement("time-taken", f, t)
        tm = influxdb_datalogger.TagMap.build(t, "test")
        datalogger = influxdb_datalogger.DataLogger(DbWriter())
        with datalogger.measure_block(m, f, tm):
            logger.info(f"Taking measurement {m}")

        assert m in datalogger._dataset, f"Measurement {m} not in the dataset"
        self.assert_data_file_doesnt_exist()

        datalogger.write_data()
        loaded_dataset = self.load_dataset_from_data_file()
        logger.info(loaded_dataset)
        assert len(loaded_dataset) == 1, f"Unexpected: {loaded_dataset}"

    def test_measure_wrapper(self):
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        m = influxdb_datalogger.Measurement("time-taken", f, t)
        tm = influxdb_datalogger.TagMap.build(t, "test")
        datalogger = influxdb_datalogger.DataLogger(DbWriter())

        def runner(arg1):
            logger.info(f"Taking measurement {m}: {arg1}")

        datalogger.measure(func=runner,
                           measurement=m,
                           field=f,
                           tags_map=tm,
                           logger=logger,
                           log_start=f"Starting {runner.__name__}",
                           log_end=f"Finished {runner.__name__}",
                           args=("test",))

        assert m in datalogger._dataset, f"Measurement {m} not in the dataset"

    def test_post_processing(self):
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        m = influxdb_datalogger.Measurement("time-taken", f, t)
        tm = influxdb_datalogger.TagMap.build(t, "test")
        datalogger = influxdb_datalogger.DataLogger(DbWriter())
        epoch = time.time()

        @influxdb_datalogger.PostProcessing.register(m)
        def validate_post_processing(datalogger: influxdb_datalogger.DataLogger, **kwargs):
            logger.info(datalogger)
            logger.info(kwargs)
            datalogger.log(m, influxdb_datalogger.FieldMap.build(f, epoch), influxdb_datalogger.TagMap.build(t, "post"))

        datalogger.add_post_processing(validate_post_processing)

        with datalogger.measure_block(m, f, tm):
            logger.info(f"Taking measurement {m}")

        self.assert_data_file_doesnt_exist()

        datalogger.write_data()
        loaded_dataset = self.load_dataset_from_data_file()
        logger.info(loaded_dataset)
        assert len(loaded_dataset) == 2, f"Unexpected: {loaded_dataset}"
        assert loaded_dataset[0]["measurement"] == m, f"Unexpected: {loaded_dataset}"
        assert loaded_dataset[1]["measurement"] == m, f"Unexpected: {loaded_dataset}"
        assert loaded_dataset[1]["fields"][f] == epoch, f"Unexpected: {loaded_dataset}"

    def test_continuous_write(self):
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        m = influxdb_datalogger.Measurement("time-taken", f, t)
        datalogger = influxdb_datalogger.DataLogger(DbWriter(), continuous_write=True)

        with datalogger.measure_block(m, f, influxdb_datalogger.TagMap.build(t, "first")):
            logger.info(f"Taking first measurement and doing an immediate write {m}")

        with datalogger.measure_block(m, f, influxdb_datalogger.TagMap.build(t, "second")):
            logger.info(f"Taking second measurement and doing an immediate write {m}")

        # NOTE: We do a continuous write, which means there's nothing in the dataset because
        # the dataset is wiped when writing the data since we don't want to write duplicates

        loaded_dataset = self.load_dataset_from_data_file()
        logger.info(loaded_dataset)
        assert len(loaded_dataset) == 2, f"Unexpected: {loaded_dataset}"
        assert loaded_dataset[0]["tags"][t] == "first", f"Unexpected: {loaded_dataset}"
        assert loaded_dataset[1]["tags"][t] == "second", f"Unexpected: {loaded_dataset}"

    def test_events(self):
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        h = influxdb_datalogger.Tag("hostname")

        m = influxdb_datalogger.Measurement("time-taken", f, t)
        etm = influxdb_datalogger.TagMap.build(h, socket.gethostname())

        datalogger = influxdb_datalogger.DataLogger(DbWriter(), GrafanaEventWriter())

        with datalogger.measure_block(m, f, influxdb_datalogger.TagMap.build(t, "first"), event_tags=etm):
            logger.info(f"Taking first measurement and logging events {m}")

        with datalogger.measure_block(m, f, influxdb_datalogger.TagMap.build(t, "second"), event_tags=etm):
            logger.info(f"Taking second measurement and logging events {m}")

        with datalogger.measure_block(m, f, influxdb_datalogger.TagMap.build(t, "third")):
            logger.info(f"Taking third measurement and NOT logging events {m}")

        self.assert_event_file_doesnt_exist()

        datalogger.write_data()

        loaded_events = self.load_events_from_event_file()
        logger.info(loaded_events)
        assert len(loaded_events) == 2

    def test_events_continuous_write(self):
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        h = influxdb_datalogger.Tag("hostname")

        m = influxdb_datalogger.Measurement("time-taken", f, t)
        etm = influxdb_datalogger.TagMap.build(h, socket.gethostname())

        datalogger = influxdb_datalogger.DataLogger(DbWriter(), GrafanaEventWriter(), continuous_write=True)

        with datalogger.measure_block(m, f, influxdb_datalogger.TagMap.build(t, "first"), event_tags=etm):
            logger.info(f"Taking first measurement and logging events {m}")

        loaded_events = self.load_events_from_event_file()
        assert len(loaded_events) == 1

        with datalogger.measure_block(m, f, influxdb_datalogger.TagMap.build(t, "second"), event_tags=etm):
            logger.info(f"Taking second measurement and logging events {m}")

        loaded_events = self.load_events_from_event_file()
        assert len(loaded_events) == 2

        with datalogger.measure_block(m, f, influxdb_datalogger.TagMap.build(t, "third")):
            logger.info(f"Taking third measurement and NOT logging events {m}")

        loaded_events = self.load_events_from_event_file()
        assert len(loaded_events) == 2

    def test_events_continuous_write_measure(self):
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        h = influxdb_datalogger.Tag("hostname")

        m = influxdb_datalogger.Measurement("time-taken", f, t)
        etm = influxdb_datalogger.TagMap.build(h, socket.gethostname())

        datalogger = influxdb_datalogger.DataLogger(DbWriter(), GrafanaEventWriter(), continuous_write=True)

        def runner(arg1):
            logger.info(f"Taking measurement {m}: {arg1}")

        datalogger.measure(func=runner,
                           measurement=m,
                           field=f,
                           tags_map=influxdb_datalogger.TagMap.build(t, "first"),
                           logger=logger,
                           log_start=f"Starting {runner.__name__}",
                           log_end=f"Finished {runner.__name__}",
                           event_tags=etm,
                           args=("first",))

        loaded_events = self.load_events_from_event_file()
        assert len(loaded_events) == 1

        datalogger.measure(func=runner,
                           measurement=m,
                           field=f,
                           tags_map=influxdb_datalogger.TagMap.build(t, "second"),
                           logger=logger,
                           log_start=f"Starting {runner.__name__}",
                           log_end=f"Finished {runner.__name__}",
                           event_tags=etm,
                           args=("second",))

        loaded_events = self.load_events_from_event_file()
        assert len(loaded_events) == 2

        datalogger.measure(func=runner,
                           measurement=m,
                           field=f,
                           tags_map=influxdb_datalogger.TagMap.build(t, "third"),
                           logger=logger,
                           log_start=f"Starting {runner.__name__}",
                           log_end=f"Finished {runner.__name__}",
                           args=("third",))

        loaded_events = self.load_events_from_event_file()
        assert len(loaded_events) == 2

    def test_default_tags(self):
        def get_hostname():
            return socket.gethostname()
        f = influxdb_datalogger.Field("duration")
        h = influxdb_datalogger.Tag("hostname", func=get_hostname)
        m = influxdb_datalogger.Measurement("time-taken", f, h)

        datalogger = influxdb_datalogger.DataLogger(DbWriter(), GrafanaEventWriter())

        with datalogger.measure_block(m, f):
            logger.info(f"Should log a default value for tag {h}")

        datalogger.write_data()

        dataset = self.load_dataset_from_data_file()
        logger.info(dataset)
        assert h in dataset[0]["tags"]

    def test_default_fields(self):
        def get_hostname():
            return socket.gethostname()
        f = influxdb_datalogger.Field("duration")
        h = influxdb_datalogger.Field("hostname", func=get_hostname)
        m = influxdb_datalogger.Measurement("time-taken", f, h)

        datalogger = influxdb_datalogger.DataLogger(DbWriter(), GrafanaEventWriter())

        with datalogger.measure_block(m, f):
            logger.info(f"Should log a default value for field {h}")

        datalogger.write_data()

        dataset = self.load_dataset_from_data_file()
        logger.info(dataset)
        assert h in dataset[0]["fields"]

    def test_tag_map_copy(self):
        t = time.time()
        time_tag = influxdb_datalogger.Tag("time")
        tm = influxdb_datalogger.TagMap.build(influxdb_datalogger.Tag("host"), "https://host.com")
        tm_copy = tm.copy()
        assert isinstance(tm_copy, influxdb_datalogger.TagMap)
        tm_copy.add(time_tag, t)
        assert time_tag not in tm

    def test_field_map_copy(self):
        t = time.time()
        time_field = influxdb_datalogger.Field("time")
        fm = influxdb_datalogger.FieldMap.build(influxdb_datalogger.Field("duration"), 3.14)
        fm_copy = fm.copy()
        assert isinstance(fm_copy, influxdb_datalogger.FieldMap)
        fm_copy.add(time_field, t)
        assert time_field not in fm

    def test_block_measurement_with_logging(self):

        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        m = influxdb_datalogger.Measurement("time-taken", f, t)
        tm = influxdb_datalogger.TagMap.build(t, "test")
        datalogger = influxdb_datalogger.DataLogger(DbWriter())
        with datalogger.measure_block(m, f, tm, logger=logger, log_start="Starting", log_end="Done"):
            logger.info(f"Taking measurement {m}")

        assert m in datalogger._dataset, f"Measurement {m} not in the dataset"

        datalogger = influxdb_datalogger.DataLogger(DbWriter())
        with influxdb_datalogger.BlockMeasurer(datalogger, m, f, tm, logger=logger, log_start="Starting", log_end="Done"):
            logger.info("Taking measurement again")

    def test_last_data_entry(self):
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        m = influxdb_datalogger.Measurement("time-taken", f, t)
        tm_1 = influxdb_datalogger.TagMap.build(t, "test_1")
        tm_2 = influxdb_datalogger.TagMap.build(t, "test_2")
        datalogger = influxdb_datalogger.DataLogger(DbWriter())
        with datalogger.measure_block(m, f, tm_1, logger=logger, log_start="Starting", log_end="Done"):
            logger.info(f"Taking measurement {m}")

        with datalogger.measure_block(m, f, tm_2, logger=logger, log_start="Starting", log_end="Done"):
            logger.info(f"Taking measurement {m}")

        last: influxdb_datalogger.DataEntry = datalogger.last_data_entry
        assert last.measurement == m
        assert last.tags == tm_2
        assert f.field_name in last.fields

    def test_last_event(self):
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        m = influxdb_datalogger.Measurement("time-taken", f, t)
        tm_1 = influxdb_datalogger.TagMap.build(t, "test_1")
        tm_2 = influxdb_datalogger.TagMap.build(t, "test_2")
        datalogger = influxdb_datalogger.DataLogger(DbWriter(), GrafanaEventWriter())
        with datalogger.measure_block(m, f, tm_1, logger=logger, log_start="Starting", log_end="Done", event_tags=influxdb_datalogger.TagMap.build(t, "event-1")):
            logger.info(f"Taking measurement {m}")

        with datalogger.measure_block(m, f, tm_2, logger=logger, log_start="Starting", log_end="Done", event_tags=influxdb_datalogger.TagMap.build(t, "event-2")):
            logger.info(f"Taking measurement {m}")

        last: influxdb_datalogger.Event = datalogger.last_event
        assert last.measurement == m
        assert last.tags_map[t] == "event-2"

    def test_descriptions(self):
        f = influxdb_datalogger.Field("duration", desc="Field used for the measurement")
        t = influxdb_datalogger.Tag("identifier", desc="Tag used for the measurement")
        m = influxdb_datalogger.Measurement("time-taken", f, t, desc="Measurement used for the test")

        assert f.desc == "Field used for the measurement", f"Field description is {f.desc}"
        assert t.desc == "Tag used for the measurement", f"Tag description is {t.desc}"
        assert m.desc == "Measurement used for the test", f"Description is {m.desc}"

    def test_time_override(self):
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        m = influxdb_datalogger.Measurement("time-taken", f, t)
        tm_1 = influxdb_datalogger.TagMap.build(t, "test_1")
        datalogger = influxdb_datalogger.DataLogger(DbWriter())
        now = datetime.datetime.utcnow()
        datalogger.log(m, influxdb_datalogger.FieldMap.build(f, 1.0), tm_1, time=now)
        assert datalogger.last_data_entry.time == str(now), f"Time is {datalogger.last_data_entry.time}"

        datalogger.log(m, influxdb_datalogger.FieldMap.build(f, 2.0), tm_1)
        logger.info(f"Time is {datalogger.last_data_entry.time}")
        assert datalogger.last_data_entry.time != str(now) and datetime.datetime.fromisoformat(datalogger.last_data_entry.time), f"Time is {datalogger.last_data_entry.time}"



from .datalogger import DataLogger, DatabaseWriter, PostProcessing, DataEntry, Dataset, Event, Events, BlockMeasurer
from .measurement import Measurement
from .tag import Tag, TagMap
from .field import Field, FieldMap



from __future__ import annotations

import datetime
import functools
import json
import logging

from abc import ABC, abstractmethod
from typing import *

from .measurement import Measurement
from .field import Field, FieldMap
from .tag import Tag, TagMap
from .utils import threaded_execution


class DatabaseWriter(ABC):

    @abstractmethod
    def write_data(self, datalogger: DataLogger):
        """Meant to write the actual data to the database being used"""
        pass


class PostProcessing:
    @staticmethod
    def register(*measurements: Measurement):
        """
        Decorator to define a postprocessing function for CSV logging as a single-entry postprocessing function.

        Function signature example:
            def postprocessor(datalogger: DataLogger, **kwargs: Dict[Measurement, List[DataEntry]])
        """

        def wrap_outer(func):
            func.__setattr__("post_processing", True)
            func.__setattr__("measurements", measurements)

            @functools.wraps(func)
            def wrap_inner(*args, **kwargs):
                return func(*args, **kwargs)

            return wrap_inner

        return wrap_outer


class PostProcessingHandler:
    def __init__(self, func: Callable[PostProcessing]):
        self.func = func
        self.measurements = func.measurements
        self.post_processing = func.post_processing


class DataEntry(dict):
    def __init__(self, measurement: Measurement, fields_map: FieldMap[Field, Any], tags_map: TagMap[Tag, Any] = None, time: datetime.datetime = None):
        fields_map = fields_map or FieldMap()
        tags_map = tags_map or TagMap()

        super(DataEntry, self).__init__()

        self.measurement: Measurement = measurement
        self.fields: FieldMap = fields_map
        self.tags: TagMap = tags_map
        self.time: str = str(time or datetime.datetime.utcnow())

    def __setattr__(self, key, value):
        super(DataEntry, self).__setattr__(key, value)
        super(DataEntry, self).__setitem__(key, value)


class Dataset(dict):
    def __init__(self):
        super(Dataset, self).__init__()

    def log(self, measurement: Measurement, fields_map: FieldMap[Field, Any], tags_map: TagMap[Tag, Any] = None, time: datetime.datetime = None) -> None:
        self.append(DataEntry(measurement, fields_map, tags_map, time=time))

    def append(self, __object: DataEntry) -> None:
        if __object.measurement not in self:
            self[__object.measurement] = list()
        self[__object.measurement].append(__object)

    def get_entries_for_measurement(self, measurement: Measurement) -> List[DataEntry]:
        return self.get(measurement)

    def as_list(self) -> List[DataEntry]:
        data = list()
        for values in self.values():
            assert len([v for v in values if isinstance(v, DataEntry)]) == len(values), f"Not all entries are of type {DataEntry}. This is not permitted"
            data.extend(values)
        return data


class Event(dict):
    def __init__(self, measurement: Measurement, start: float, stop: float, tags_map: TagMap):
        super(Event, self).__init__()
        self.measurement = measurement
        self.start = start
        self.stop = stop
        self.tags_map = tags_map

    def __setattr__(self, key, value):
        super(Event, self).__setattr__(key, value)
        super(Event, self).__setitem__(key, value)


class Events(dict):
    def __init__(self):
        super(Events, self).__init__()

    def log(self, measurement: Measurement, start: float, stop: float, tags_map: TagMap):
        self.append(Event(measurement, start, stop, tags_map))

    def append(self, __object: Event) -> None:
        if __object.measurement not in self:
            self[__object.measurement] = list()
        self[__object.measurement].append(__object)

    def get_entries_for_measurement(self, measurement: Measurement) -> List[Event]:
        return self.get(measurement)

    def as_list(self) -> List[Event]:
        data = list()
        for values in self.values():
            assert len([v for v in values if isinstance(v, Event)]) == len(values), f"Not all entries are of type {Event}. This is not permitted"
            data.extend(values)
        return data


class BlockMeasurer:
    def __init__(self,
                 datalogger: DataLogger,
                 measurement: Measurement,
                 field: Field,
                 tags_map: TagMap = None,
                 event_tags: TagMap = None,
                 logger: logging.Logger = None,
                 log_lvl: int = logging.INFO,
                 log_start: str = None,
                 log_end: str = None):
        f"""
        Create a BlockMeasure object to measure the time it takes to run an arbitrary block of code 
        
        Args:
            datalogger: The datalogger object to attach this BlockMeasurer object to. 
            measurement: The main measurement to log the time to.
            field: The field in the main measurement to log the duration to.
            tags_map: Optional map of tags to assign to the measurement.
            event_tags: Optional map of tags used for logging events. Events will be written if and only if this variable isn't None.
            logger: Optional {logging.Logger} object to write log messages. If this is passed but {log_start} and/or {log_end} are not, something generic will be written.
            log_lvl: Optional logging level for writing log messages using {logger}. Default is {logging.INFO}. 
            log_start: Optional logging message to write before the block is executed and when the measurement is started. Will be something generic if {logger} is passed and this is left as {None}. 
            log_end: Optional logging message to write after the block is executed and when the measurement is stopped. Will be something generic if {logger} is passed and this is left as {None}. 
        """
        self.datalogger = datalogger
        self.measurement = measurement
        self.field: Field = field
        self.tags_map: TagMap = tags_map
        self.start_time: datetime.datetime = None
        self.end_time: datetime.datetime = None
        self.event_tags: TagMap = event_tags
        self.logger = logger
        self.log_lvl = log_lvl
        self.log_start = log_start or f"Starting block-measurement for {measurement}"
        self.log_end = log_end or f"Stopping block-measurement for {measurement}"

    def __enter__(self):
        if self.logger:
            self.logger.log(self.log_lvl, self.log_start)
        self.start_time = datetime.datetime.utcnow()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_time = datetime.datetime.utcnow()
        duration = (self.end_time - self.start_time).total_seconds()
        if self.event_tags:
            self.datalogger.log_event(self.measurement, self.start_time, self.end_time, self.event_tags)

        self.datalogger.log(self.measurement, FieldMap.build(self.field, duration), self.tags_map)

        if self.logger:
            self.logger.log(self.log_lvl, self.log_end)


class DataLogger:
    def __init__(self, *database_writers: DatabaseWriter, continuous_write: bool = False):
        f"""
        Args:
            database_writers: An arbitrary number of database writers created by the user that are of the type {DatabaseWriter} 
            continuous_write: A bool to tell if writes should be done immediately when {self.log.__name__} is invoked. Removes the need to use {self.write_data.__name__}`
        """
        self._dataset: Dataset = Dataset()
        self.postprocessing_functions = list()
        self._db_writers = list()
        self.continuous_write = continuous_write
        self._events: Events = Events()
        for database_writer in database_writers:
            assert isinstance(database_writer, DatabaseWriter), \
                f"A database writer passed to a '{DataLogger.__module__}.{DataLogger}' object must inherit from type '{DatabaseWriter}'."

            self._db_writers.append(database_writer)

    def __str__(self):
        return json.dumps(self._dataset, indent=2)

    @property
    def last_data_entry(self) -> DataEntry:
        return self.dataset[-1]

    @property
    def last_event(self) -> Event:
        return self.events[-1]

    @property
    def dataset(self) -> List[DataEntry]:
        return self._dataset.as_list()

    @property
    def events(self) -> List[Event]:
        return self._events.as_list()

    @property
    def db_writers(self):
        return self._db_writers

    def add_post_processing(self, func: Callable[PostProcessing]):
        """
        Adds a post-processing function to a data logger.

        Args:
            func: The callable function to add as a post-processing method.
        """
        assert hasattr(func, "post_processing") and func.post_processing, f"This function does not appear to be registered as a post-processing method. Please use decorator '@{PostProcessing.register}' to register it"
        self.postprocessing_functions.append(PostProcessingHandler(func))

    def measure(self,
                func: Callable,
                measurement: Measurement,
                field: Field,
                tags_map: TagMap = None,
                event_tags: TagMap = None,
                logger: logging.Logger = None, log_start: str = None, log_end: str = None,
                args: tuple = None, kwargs: dict = None):
        """
        This function wraps any arbitrary function and times the duration for that function and then logs the duration
        to a DataLogger object for a given measurement and a field name.
        Wrapping function to call an arbitrary function and calculate the duration of that function.

        Args:
            func: The function to call and calculate the time for.
            measurement: The main measurement to log the time to.
            field: The field in the main measurement to log the duration to.
            tags_map: Optional map of tags to assign to the measurement.
            event_tags: Optional map of tags used for logging events. Events will be written if and only if this variable isn't None.
            logger: Optional logger object. Used to log a message before and/or after calling the function.
            log_start: Optional log message to log before calling the function. Requires that the argument logger is passed to this function.
            log_end: Optional log message to log after calling the function. Requires that the argument logger is passed to this function.
            args: A tuple of arguments to pass to func.
            kwargs: A dict of keyword-arguments to pass to func.
        """
        #
        # Verify the input.
        #
        assert isinstance(func, Callable), f"'func' must be a callable object."

        if args is None:
            args = tuple()
        else:
            assert isinstance(args, tuple), "args must be passed as a tuple"
        if kwargs is None:
            kwargs = dict()
        else:
            assert isinstance(kwargs, dict), "kwargs must be passed as a dict"

        if log_start is not None or log_end is not None:
            assert logger is not None and isinstance(logger, logging.Logger), f"If either 'log_start' or 'log_end' is passed to this method, 'logger' must also be passed as a '{logging.Logger}' object"

        if log_start:
            logger.info(log_start)

        #
        # Execute wrapped function
        #
        ts_start = datetime.datetime.utcnow()
        return_values = func(*args, **kwargs)
        ts_stop = datetime.datetime.utcnow()

        #
        # Get duration
        #
        dur = (ts_stop - ts_start).total_seconds()

        #
        # Log event for measurement (if event_tags is not None)
        #
        if event_tags:
            self.log_event(measurement, ts_start, ts_stop, event_tags)

        #
        # Log duration to measurement.
        #
        self.log(measurement, FieldMap.build(field, dur), tags_map=tags_map)

        if log_end:
            logger.info(log_end)

        return return_values

    def measure_block(self, measurement: Measurement,
                      field: Field,
                      tags_map: TagMap = None,
                      event_tags: TagMap = None,
                      logger: logging.Logger = None,
                      log_lvl: int = logging.INFO,
                      log_start: str = None,
                      log_end: str = None):
        f"""
        This function uses the `with` syntax in Python to measure the duration for an arbitrary block of code.

        Usage:
        ```
        f = influxdb_datalogger.Field("duration")
        t = influxdb_datalogger.Tag("identifier")
        m = influxdb_datalogger.Measurement("time-taken", f, t)
        tm = influxdb_datalogger.TagMap.build(t, "test")
        datalogger = influxdb_datalogger.DataLogger()

        with datalogger.measure_block(measurement=m, field=f, tags_map=tm):
            time.sleep(1)
        ```

        Args:
            measurement: The main measurement to log the time to.
            field: The field in the main measurement to log the duration to.
            tags_map: Optional map of tags to assign to the measurement.
            event_tags: Optional map of tags used for logging events. Events will be written if and only if this variable isn't None.
            logger: Optional {logging.Logger} object to write log messages. If this is passed but {log_start} and/or {log_end} are not, something generic will be written.
            log_lvl: Optional logging level for writing log messages using {logger}. Default is {logging.INFO}. 
            log_start: Optional logging message to write before the block is executed and when the measurement is started. Will be something generic if {logger} is passed and this is left as {None}. 
            log_end: Optional logging message to write after the block is executed and when the measurement is stopped. Will be something generic if {logger} is passed and this is left as {None}. 
        """
        return BlockMeasurer(datalogger=self, measurement=measurement, field=field, tags_map=tags_map, event_tags=event_tags, logger=logger, log_lvl=log_lvl, log_start=log_start, log_end=log_end)

    def log(self, measurement: Measurement, fields_map: FieldMap, tags_map: TagMap = None, time: datetime.datetime = None):
        """
        Used to log data for a measurement using a fields_map and an optional tags_map.

        Args:
            measurement: The measurement to log the data for.
            fields_map: A FieldMap object to write data to the measurement.
            tags_map: An optional TagMap to tag the data written for the measurement.
            time: An optional datetime object to use as the timestamp for the data written. If not passed, the current time will be used (using datetime.datetime.utcnow()).
        """

        assert isinstance(measurement, Measurement), f"'measurement' must be of type '{Measurement.__module__}.{Measurement}'."
        assert isinstance(fields_map, FieldMap), f"'fields_map' must be of type '{FieldMap.__module__}.{FieldMap}', not '{type(fields_map)}'"
        assert isinstance(tags_map, TagMap) or tags_map is None, f"'tags_map' must be {None}, or of type '{TagMap.__module__}.{TagMap}', not '{type(tags_map)}'"

        tags_defaulted = measurement.get_defaulted_tags()
        tags_predefined = measurement.get_predefined_tags()
        # Add together all maps of tags
        tags_all = TagMap()
        # Add the set of defaulted tags first
        tags_all.update(tags_defaulted)
        # Add the set of pre-defined tags second
        tags_all.update(tags_predefined)
        # Add the set of configured tags third (if they aren't None) so that the user may overwrite pre-defined tags if they have been configured manually.
        if tags_map:
            tags_all.update(tags_map)
        # Assert tags are defined in measurement
        for tag in tags_all:
            assert tag in measurement.tags, f"Tag '{tag}' is not defined as a tag for measurement '{measurement}'"

        fields_defaulted = measurement.get_default_fields()
        fields_predefined = measurement.get_predefined_fields()
        # Add together all maps of fields
        fields_all = FieldMap()
        # Add the set of defaulted fields first
        fields_all.update(fields_defaulted)
        # Add the set of pre-defined fields second
        fields_all.update(fields_predefined)
        # Add the set of configured fields third so that the user may overwrite pre-defined tags if they have been configured manually.
        fields_all.update(fields_map)
        # Assert fields are in measurement
        for field in fields_all:
            assert field in measurement.fields, f"Field '{field}' is not defined as a field for measurement '{measurement}'"

        # Log the measurement to the dataset
        self._dataset.log(measurement, fields_all, tags_all, time=time)

        # If continuous_write is configured we should write the data immediately
        if self.continuous_write:
            self.write_data()

    def log_event(self, measurement: Measurement, start_time: datetime.datetime, stop_time: datetime.datetime, event_tags_map: TagMap = None):
        """
        Used to log data for a measurement using a fields_map and an optional tags_map.

        Args:
            measurement: The measurement to log the event for.
            start_time: datetime.datetime object for when the event started. Should be in UTC
            stop_time: datetime.datetime object for when the event stopped. Should be in UTC
            event_tags_map: An optional TagMap to tag the event data written for the measurement.
        """
        epoch_ts_start = start_time.replace(tzinfo=datetime.timezone.utc).timestamp()
        epoch_ts_stop = stop_time.replace(tzinfo=datetime.timezone.utc).timestamp()
        self._events.log(measurement=measurement, start=epoch_ts_start, stop=epoch_ts_stop, tags_map=event_tags_map)

    def write_data(self):
        """
        Write data to all database writers configured to the data logger.

        Steps done:
        1. Validate the integrity of the data that has been written to the logger.
        2. Run all post-processing methods configured for the logger.
        3. Since the dataset have changed after post-processing, we need to validate the dataset again.
        4. Write the dataset to all database writers configured for the logger.
        5. Clean the dataset since the data has been written.
        """
        #
        # Validate the integrity of the dataset based on how it has been created.
        #
        self._validate_dataset()

        #
        # Do post-processing that has been added on the recently validated dataset.
        #
        self._do_post_processing()

        #
        # Validate the integrity of the dataset again after doing post-processing.
        #
        self._validate_dataset()

        #
        # Write the dataset to all database writers configured to the data logger.
        #
        for writer in self._db_writers:
            assert isinstance(writer, DatabaseWriter)
            writer.write_data(self)

        #
        # Clean the dataset since it has all been written to the databases.
        #
        self._dataset.clear()
        self._events.clear()

    def _validate_dataset(self):
        def validate_entry(entry: DataEntry):
            measurement = entry.measurement
            assert isinstance(measurement, Measurement)
            measurement_tags = measurement.tags
            measurement_fields = measurement.fields
            fields = entry.fields
            tags = entry.tags

            for field in measurement_fields:
                assert field in fields and fields[field] is not None, \
                    f"Field '{field}' has not been logged with a value. This is not permitted."

            for tag in measurement_tags:
                assert (tag in tags and tags[tag] is not None) or tag in measurement.optional_tags, \
                    f"Tag '{tag}' has not been configured with a value, and is not configured as an optional tag. This is not permitted."

        threaded_execution(validate_entry, self.dataset)

    def _do_post_processing(self):
        for post_processing in self.postprocessing_functions:
            assert isinstance(post_processing, PostProcessingHandler)
            assert hasattr(post_processing.func, "post_processing") and post_processing.func.post_processing
            all_measurement_entries = dict()
            for measurement in post_processing.measurements:
                measurement_entries = self._dataset.get_entries_for_measurement(measurement)
                all_measurement_entries[measurement] = measurement_entries
            post_processing.func(self, **all_measurement_entries)



from typing import Union, Iterable

from .field import Field, FieldMap
from .tag import Tag, TagMap


class Measurement(str):
    def __new__(cls, measurement_name: str,
                *fields_and_tags: Union[Tag, Field], optional_tags: Iterable[Tag] = None, desc: str = None):
        """
        Define a measurement to be used for logging data.

        Args:
            measurement_name: The name of the measurement.
            fields: The fields to be used for the measurement.
            *tags: An arbitrary tuple of arbitrary tags to be used for tagging the data.
            optional_tags: Tags that are optional and which must not be configured.
            desc: An optional description of the measurement.
        """

        obj = super().__new__(cls, measurement_name)
        obj.key_name = measurement_name
        obj.desc = desc
        obj.fields = set()
        obj.tags = set()
        for _obj in fields_and_tags:
            if isinstance(_obj, Tag):
                obj.tags.add(_obj)
            elif isinstance(_obj, Field):
                obj.fields.add(_obj)
            else:
                raise TypeError(f"Unsupported type for field/tag: {type(obj)}")

        if optional_tags:
            for tag in optional_tags:
                assert isinstance(tag, Tag), f"Unsupported type for lenient tag: {type(tag)}. Must be type {Tag}"
                obj.tags.add(tag)

        obj.optional_tags = optional_tags if optional_tags else set()
        return obj

    def get_predefined_tags(self) -> TagMap:
        """Run 'func' for all tags that have the value defined already."""
        fm = TagMap()
        for tag in self.tags:
            assert isinstance(tag, Tag), f"Tag {tag} is of type {type(tag)}, not {Tag}"
            if tag.func is not None:
                fm[tag] = tag.func(*tag.args_func)
        return fm

    def get_defaulted_tags(self) -> TagMap:
        """Get all tags with defaults defined"""
        tm = TagMap()
        for tag in self.tags:
            assert isinstance(tag, Tag), f"Tag {tag} is of type {type(tag)}, not {Tag}"
            if tag.default is not None:
                tm[tag] = tag.default
        return tm

    def get_predefined_fields(self) -> FieldMap:
        """Run 'func' for all fields that func set."""
        fm = FieldMap()
        for field in self.fields:
            assert isinstance(field, Field), f"Field {field} is of type {type(field)}, not {Field}"
            if field.func is not None:
                fm[field] = field.func(*field.args_func)
        return fm

    def get_default_fields(self) -> FieldMap:
        """Get all tags with defaults defined"""
        fm = FieldMap()
        for field in self.fields:
            assert isinstance(field, Field), f"Field {field} is of type {type(field)}, not {Field}"
            if field.default is not None:
                fm[field] = field.default
        return fm



from __future__ import annotations

from typing import Callable, AnyStr, Any

from .datalogger_map import DataLoggerMap


class Tag(str):
    def __new__(cls, tag_name: str, *args_func: object, func: Callable = None, desc: str = None, default: object = None):
        """
        Define a Tag to be used for logging data.

        Args:
            tag_name: The name of the tag.
            args_func: Optional arguments to pass to func.
            func: An optional argument to define a function to call when a measurement using this tag is logged to the dataset.
             Using this argument will turn the tag into a tag that may be handled automatically by the datalogger.
            desc: An optional description of the tag.
            default: An optional argument for defining a default value so the value doesn't have to be logged manually.

        Raises:
            ValueError: If both func and default are not None.
        """
        if func and default:
            raise ValueError(f"Both 'func' and 'default' cannot be configured at the same time.")

        obj = super().__new__(cls, tag_name)
        obj.tag_name = tag_name
        obj.func = func
        obj.desc = desc
        obj.default = default
        if not args_func:
            args_func = tuple()
        obj.args_func = args_func
        return obj


class TagMap(DataLoggerMap):
    @property
    def key_type(self):
        return Tag

    def add(self, t: Tag, v: Any):
        assert t not in self, f"Tag {t} is already present in the tag-map (it is mapped to {self.get(t)})"
        self.update({t: v})

    def copy(self) -> TagMap[AnyStr, Any]:
        r = TagMap(super(TagMap, self).copy())
        return r



from concurrent.futures import ThreadPoolExecutor
from typing import *


def threaded_execution(target: Callable, *args_as_iterable: Union[Set, List, Tuple, KeysView, ItemsView, ValuesView], pool_size: int = None):
    """
    Function to wrap ThreadedExecutor. Takes a function and arguments as an iterable and creates
    an executor pool to run the function concurrently.
    Args:
        target: A callable that takes each value in 'args_as_iterable' as an argument.
        args_as_iterable: An iterable where arguments for 'target' is in each element in the iterable.
        pool_size: Optional argument for setting the size of the thread pool. Default is None, in which case len(args_as_iterable) will be used as pool size. If set to 0, pool_size will effectively unlimited.
    Returns:
        A list with the return values from 'target'.
    """
    class ThreadedExecutor(object):
        def __init__(self, size=10):
            self.executor = ThreadPoolExecutor(max_workers=size)
            self.map = self.executor.map

        def __enter__(self):
            def nonlazy_map(function, *iterable_args):
                return list(self.map(function, *iterable_args))
            return nonlazy_map

        def __exit__(self, exc_type, exc_val, exc_tb):
            if hasattr(self, 'executor'):
                self.executor.shutdown()

    zipped = list(zip(*args_as_iterable))
    if len(zipped) == 0:
        return

    # Set pool_size to pool_size if it was configured, otherwise use len(args_as_iterable)
    pool_size = pool_size or len(zipped)

    assert isinstance(pool_size, int)
    with ThreadedExecutor(size=pool_size) as threaded_executor_function:
        _ret = threaded_executor_function(target, *args_as_iterable)
        return _ret


from __future__ import annotations

from typing import Callable, Any, AnyStr

from .datalogger_map import DataLoggerMap


class Field(str):
    def __new__(cls, field_name: str, *args_func: object, func: Callable = None, desc: str = None, default: object = None):
        """
        Define a Field to be used for logging data.

        Args:
            field_name: The name of the field.
            args_func: Optional arguments to pass to func.
            func: An optional argument to define a function to call when a measurement using this field is logged to the dataset.
             Using this argument will turn the field into a field that may be handled automatically by the datalogger.
            desc: An optional description of the field.
            default: An optional argument for defining a default value so the value doesn't have to be logged manually.
        """

        obj = super().__new__(cls, field_name)
        obj.field_name = field_name
        obj.func = func
        obj.desc = desc
        obj.default = default

        if not args_func:
            args_func = tuple()
        obj.args_func = args_func
        return obj


class FieldMap(DataLoggerMap):
    @property
    def key_type(self):
        return Field

    def add(self, f: Field, v: Any):
        assert f not in self, f"Field {f} is already present in the field-map (it is mapped to {self.get(f)})"
        self.update({f: v})

    def copy(self) -> FieldMap[AnyStr, Any]:
        r = FieldMap(super(FieldMap, self).copy())
        return r
